//Important information regarding the configuration of the pipetting robot:
//Pin 2 - Lower delimiter of the z-axis
//Pin 3 - Upper delimiter of the z-axis
//Pin 8 - Delimiter of the x-axis (onyl one for initialization, stepper motor is used)
//Pin 11 - Button of the mutichannel-pipette (hard-wired to its PCB)
//Pin 14 - Output relay module which controls the power-supply of the mutli-channel pipette  (HIGH switches pipette off!)
//Pin 15 - Output relay module for control of the magnet valve (HIGH switches the flow onto the plate!)
//Pins 16, 17, 18, 19: Digital in defined as pull-up, i.e. they are HIGH by default and LOW as soon as swith is pressed!
//Distance between delimiter of the x-axis and row 1 on the multiwell platte: 430 steps (depends on your setup)
//Distance between individual rows on the multiwell plate: 92 steps (depends on your setup)

// Includes
#include <Wire.h>
#include <Adafruit_MCP23017.h>
#include <Adafruit_RGBLCDShield.h>
#include <Adafruit_MotorShield.h>
#include "utility/Adafruit_PWMServoDriver.h"

// Initialize LCD shield
Adafruit_RGBLCDShield lcd = Adafruit_RGBLCDShield();
#define RED 0x1
#define YELLOW 0x3
#define GREEN 0x2
#define TEAL 0x6
#define BLUE 0x4
#define VIOLET 0x5
#define WHITE 0x7 
#define OFF 0x0


//Initialize motor shields
Adafruit_MotorShield AFMSbot(0x61); // Rightmost jumper closed, bottom board
Adafruit_MotorShield AFMStop(0x60); // Default address, no jumpers, top board

// Initialize Motors
Adafruit_DCMotor *myMotor1 = AFMStop.getMotor(1); //DC motor on a top shield
Adafruit_StepperMotor *myStepper1 = AFMSbot.getStepper(200, 1); //Stepper motor on bottom shield

// Define Variabels for FET

int long_delay = 1000;
int max_row = 5; //Counted from 0, this results in 6 rows!
int probe_positon_from_xzero = 40;
int probe_abfall_from_xzero = 150;
int row_distance_backward = 93;
int row_distance_forward = 93;


boolean should_run = true;
boolean stop_detected = false;
boolean go = false;  


long incubation_time = 0;


//######CUSTOM PROCEDURE DECLARATIONS#####

//Home all axes procedure declaration
void home_all_axes () {

  //Move up until upper z end stop is reached
  while(digitalRead(3) == HIGH){
    myMotor1->run(FORWARD);
    delay(5);
    myMotor1->run(RELEASE);     
  }

  //Move until X-axis end stop is reached
  while(digitalRead(8) == HIGH){
    myStepper1->onestep(FORWARD, SINGLE);     
  }

  myStepper1->step(415, BACKWARD, SINGLE);


}

void serial_dilution () {


  print_debug("SERIAL DILUTION...");

  //Take sample
  //Move until lower endstop is reached
  while(digitalRead(2) == HIGH){
    myMotor1->run(BACKWARD);
    delay(5);
    myMotor1->run(RELEASE);
  }
  //Pipet  
  digitalWrite(11, HIGH);
  delay(10);
  digitalWrite(11, LOW);
  delay(2500);


  //Move up until upper end stop is reached
  while(digitalRead(3) == HIGH){
    myMotor1->run(FORWARD);
    delay(5);
    myMotor1->run(RELEASE);    
  }
  
  delay(20);
  
  //Do dilution
  int i = 0;
  while (i < 5) {
    myStepper1->step(95, BACKWARD, SINGLE);
    //Move until lower endstop is reached  
    while(digitalRead(2) == HIGH){
      myMotor1->run(BACKWARD);
      delay(5);
      myMotor1->run(RELEASE);
    }
  
    //Pipet
    int j = 0;
    while (j < 4) {
      digitalWrite(11, HIGH);
      delay(10);
      digitalWrite(11, LOW);
      delay(2500);
      j++;
    }
  
    //Move up until upper end stop is reached
    while(digitalRead(3) == HIGH){
      myMotor1->run(FORWARD);
      delay(5);
      myMotor1->run(RELEASE);     
    }
    i++;
  }
  //Discard last fill
  myStepper1->step(690, FORWARD, SINGLE);
  
  //Move until lower endstop is reached  
  while(digitalRead(2) == HIGH){
    myMotor1->run(BACKWARD);
    delay(5);
    myMotor1->run(RELEASE);
  }
  
  digitalWrite(11, HIGH);
  delay(10);
  digitalWrite(11, LOW);
  delay(2500);
  
  //Move up until upper end stop is reached
  while(digitalRead(3) == HIGH){
    myMotor1->run(FORWARD);
    delay(5);
    myMotor1->run(RELEASE);     
  }
}

//#####END CUSTOM PROCEDURE DECLARATIONS#####

//##### PROCEDURE#####
void setup () {
  // Debugging output
  Serial.begin(9600);
  // set up the LCD's number of columns and rows: 
  lcd.begin(16, 2);

  // Home Screen
  lcd.print("SERIAL DILUTION");
  lcd.setCursor(0, 1);
  lcd.print("BUDDY V1.1");
  lcd.setBacklight(WHITE);
  
  delay(1000);


  //Start shields and motors
  AFMSbot.begin(); // Start the bottom shield
  AFMStop.begin(); // Start the top shield
  myMotor1->setSpeed(200); //Start DC motor
  myMotor1->run(RELEASE);
  
  // pins
  pinMode(2, INPUT_PULLUP); //z-axis, lower end stop
  pinMode(3, INPUT_PULLUP); //z-axis, upper end stop
  pinMode(8, INPUT_PULLUP); //x-axis endstop
  
  pinMode(11, OUTPUT);
  pinMode(14, OUTPUT);  
  pinMode(15, OUTPUT);
  
  pinMode(16, INPUT_PULLUP);
  pinMode(17, INPUT_PULLUP);
  pinMode(18, INPUT_PULLUP);
  pinMode(19, INPUT_PULLUP);
  
  //Home all axes
  //home_all_axes();
  move_up();
  move_xaxis_zero();
  move_to_abfall();
  print_debug("HOME POSITON");

  
}

//#####END  PROCEDURE#####

//#####MAIN PROGRAM LOOP#####


void checkforstop(){
   uint8_t buttons = lcd.readButtons();

   if (buttons & BUTTON_DOWN) {      
      should_run = false;
      stop_detected = true;
      Serial.write("STOP DETECTED");      

      if (stop_detected)
      {
        
        print_debug("RUN PAUSED...");        
        should_run = true;

        lcd.setCursor(0, 1);
     


        lcd.print("PRESS LEFT TO CONTINUE");
        lcd.scrollDisplayRight();

        do 
        {                    
             uint8_t buttons = lcd.readButtons();
             if (buttons & BUTTON_LEFT) {     
                go = true;
             }

        } while ( go !=true);

        go = false;

      }

   }
}



void move_up (){
  
    //Move up until upper end stop is reached
  while(digitalRead(3) == HIGH){
    myMotor1->run(FORWARD);
    delay(5);
    myMotor1->run(RELEASE);    
  }
  delay(20);

  
  
}

void print_debug(String message){
  lcd.clear();
  lcd.print(message);
}


void print_debug_row(String message){
  lcd.clear();
  lcd.print(message);
}



void move_down(){

    while(digitalRead(2) == HIGH){
    myMotor1->run(BACKWARD);
    delay(5);
    myMotor1->run(RELEASE);
  }
 
  
}

void move_xaxis_zero(){
 
  checkforstop();

  if(should_run){

    print_debug("RESET X AXIS");
    while(digitalRead(8) == HIGH){

      myStepper1->onestep(FORWARD, SINGLE);     
    }

    delay(1000);

  }
  
}


void move_to_probe(){
  //Move until X-axis end stop is reached

  checkforstop();

  if (should_run){

    move_xaxis_zero();


    checkforstop();

    if(should_run){

      myStepper1->step(probe_positon_from_xzero, BACKWARD, SINGLE);
      delay(1000);   
    }

  
  }
   
   
}

void move_to_abfall(){
  //Move until X-axis end stop is reached

  checkforstop();

  if(should_run)
  {
   myStepper1->step(probe_abfall_from_xzero, BACKWARD, SINGLE);
   delay(1000); 
  }
   
   
   
}

void ziehe_probe(){

  
   move_up();
   move_to_probe();
   delay(long_delay);
   move_down();
   take_sample();  
   move_up();
}


void bestuecke(int step_position){


  checkforstop();

  if(should_run)
  {

     print_debug("EQUIP SAMPLE...");

    myStepper1->step(step_position, BACKWARD, SINGLE);


    checkforstop();
    if(should_run){
        move_up();      
    }


    checkforstop();
    if(should_run){
      move_down();          
    }

    //Pipet  

    digitalWrite(11, HIGH);
    delay(10);
    digitalWrite(11, LOW);
    delay(3500);


    checkforstop();
    if(should_run){
          move_up();
    }


  }

  
}


void take_sample() {

  checkforstop();

  if (should_run)
  {

      print_debug("TAKE SAMPLE...");

       while(digitalRead(2) == HIGH){
          myMotor1->run(BACKWARD);
          delay(5);
          myMotor1->run(RELEASE);
        }

        //Pipet  
        digitalWrite(11, HIGH);
        delay(10);
        digitalWrite(11, LOW);
        delay(3500);
        
        move_up();

  }
 

}


void bestuecke_alle_reihen(){

  checkforstop();

  if (should_run)
  {

    move_xaxis_zero();
    int step_counter = 845;

    for (int i=0; i <= max_row; i++){
       String counter = String(i+1);       
       ziehe_probe();
       delay(1000);
       bestuecke(step_counter);      
       step_counter = step_counter - row_distance_backward;
    }


    checkforstop();

    if(should_run){

      myStepper1->step(260, FORWARD, SINGLE);    

    }

  
  }

  
}


void pick_garbage(int step_position){


   checkforstop();

   if(should_run )
   {

    print_debug("DISPOSE SAMPLE...");
    move_up();  


    checkforstop();

    if(should_run){

      myStepper1->step(step_position, BACKWARD, SINGLE);    

      move_down();
      take_sample();
      move_up();

    }

   }

  
}


void drop_pipet() {    

    checkforstop();

    if(should_run){

        move_down();
        digitalWrite(11, HIGH);
        delay(10);
        digitalWrite(11, LOW);
        delay(1000);
        move_up();

    }

    
}


void clean_rows(){


  checkforstop();


  //move_xaxis_zero();
  //move_to_abfall();


  if(should_run){

    int step_counter = 735;
  for (int i=0;i <= max_row; i++)
  {   
    String counter = String(i+1);
    print_debug("ROW +counter");
    pick_garbage(step_counter);
    delay(1000);



    checkforstop();

    if(should_run){

        myStepper1->step(step_counter, FORWARD, SINGLE);
        drop_pipet();
        delay(1000);
        step_counter = step_counter - row_distance_backward;
    }


    
  }

  }

  
}


void single_run_clean_rows(){

  move_xaxis_zero();
  move_to_abfall();
  clean_rows();
  myStepper1->step(110, FORWARD, SINGLE);

}


void full_run() {
  move_xaxis_zero();
  move_to_abfall();
  //Switch flow onto the plate
  digitalWrite(15, HIGH);
  
  //Reset pipette
  digitalWrite(14, HIGH);
  delay(2000);
  digitalWrite(14, LOW);
  delay(40000);
  
  //Clean rows
  clean_rows();
  
  //Wait 1 min for complete exchange of liquid on the plate
  delay(60000);
  
  //Reset pipette
  digitalWrite(14, HIGH);
  delay(2000);
  digitalWrite(14, LOW);
  delay(40000);
  
  //Pipet into all rows
  bestuecke_alle_reihen();
  
  //Swith flow into the recycling mode
  digitalWrite(15, LOW);
  
  print_debug("INCUBATION");



  //660
  for(int i = incubation_time; i >=0; i--)
  {   
      String counter = String(i);         
      lcd.clear();
      lcd.print("INCUBATION");
      lcd.setCursor(0, 1);
      lcd.print(counter);

      if (i == 9)
      {
        lcd.clear();
        lcd.print("INCUBATION");
        lcd.setCursor(0, 1);
        lcd.print(counter);

      }

      delay(1000);
  }  

  
}



void pipet(){

  digitalWrite(11, HIGH);
  delay(10);
  digitalWrite(11, LOW);
    
}


void testrun(){
  clean_rows();
  bestuecke_alle_reihen();  
}



void set_incubation_time()
{

  int incubation_time_minutes = 1;

  print_debug ("Choose Incubation Time");
  lcd.setCursor(0,1);
  lcd.print(incubation_time_minutes);
  boolean time_selected = false;


  do 
  {
      uint8_t buttons = lcd.readButtons();
        
        if (buttons & BUTTON_UP) {                      
           delay(500); 
           incubation_time_minutes++;
                     
           lcd.clear();
           lcd.print(incubation_time_minutes);
       }

       if (buttons & BUTTON_DOWN) { 
            
          delay(500);
          incubation_time_minutes--;                
          lcd.clear();
          lcd.print(incubation_time_minutes);                   
       }

        if (buttons & BUTTON_SELECT) {
        
          incubation_time = incubation_time_minutes * 60;
          lcd.clear();
          lcd.print(incubation_time);          
          time_selected = true;

        }

  } while ( time_selected != true );


}



void loop() {

  uint8_t buttons = lcd.readButtons();

  if (buttons) {
    if (buttons & BUTTON_LEFT) {

        set_incubation_time();

        while ( true)
        {
         full_run();        
        }
       
    }   

    if (buttons & BUTTON_RIGHT) {  

        home_all_axes();    
        serial_dilution ();
    }

    

  }
}

//####END MAIN PROGRAM LOOP#####


